/*
 * File...... opticspy_tx.c
 * Author.... Joe Grand, Grand Idea Studio, Inc. [@joegrand]
 * Updated... 6 Nov 2017
 *
 * This program uses the WAN LED on the TP-Link TL-WR841N as an optical covert channel.
 * Optical covert channels transmit data by modulating visible light in a way that is 
 * undetectable to the human eye. ASCII printable data is sent via the LED with NRZ 
 * (Non-Return-to-Zero) encoding to emulate a standard UART interface.
 *
 * More details about this example and the hardware needed to receive the
 * data can be found at:
 * www.grandideastudio.com/portfolio/optical-covert-channels/
 *
 * Code based on "Bit Bang GPIO UART from Linux Kernel," www.ctrlinux.com/blog/?p=154
 * and "Access GPIO from Linux user space," falsinsoft.blogspot.com/2012/11/access-
 * gpio-from-linux-user-space.html
 *
 * Cross compiled from Ubuntu 16.10 using dd-wrt's toolchain-mips_24kc_gcc-7.2.0_musl
 * $ mips-linux-uclibc-gcc opticspy_tx.c -o opticspy
 * 
 * TP-Link TL-WR841N v11.0
 * Qualcomm Atheros QCA9533 ver 2 rev 1.0 (0x0160)
 * MIPS 24Kc V7.4
 * BogoMIPS: 432.53
 *
 */

#define NEED_PRINTF 	1	// Required for printf to use stdout 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

/* Constants */
#define BIT_TIME_US     416     // Bit time in uS, baud rate = 1/BIT_TIME = 2400
#define LOOP_DELAY      1       // Delay in seconds between characters
#define WAN_LED         2       // GPIO pin for the WAN LED (bicolor)

char const msg_covert[] = {"SECRET\n"};

/* Function prototypes */
void gpio_open(int *, int);
void gpio_set_value(int, int);
void gpio_close(int);
void bitbang_string(int, const char*);
void bitbang_char(int, char);
void udelay(int);

/* Core functions */
int main (void) 
{ 
    int fd;     // File descriptor for access to sysfs
    //FILE *file; // File descriptor for access to file
    //int c;      // Character read from file

    printf("Welcome to OpticSpy!\n\n");

	gpio_open(&fd, WAN_LED);
    gpio_set_value(fd, 0); // LED on during idle state (low)

    printf("Exfiltrating data via LED...\n");
    sleep(LOOP_DELAY);
    while(1)
    {
        bitbang_string(fd, msg_covert); // Send pre-defined message to LED        

        // Send the root password hash (first entry) from /etc/passwd
        /*file = fopen ("/etc/passwd", "r");
        if (file)
        {
            while ((c = getc(file)) != '\n')
            {
                bitbang_char(fd, c);
                putchar(c);                 // Display character to console
                fflush(stdout);
                sleep(LOOP_DELAY);          // Inter-character delay
            }
            fclose(file);
            bitbang_char(fd, '\n');
            putchar('\n');
        }*/
    }
 
    gpio_close(fd);
}

void gpio_open(int *fd, int gpio)
{
	char buf[64]; // Buffer for string manipulation
	
	sprintf(buf, "/sys/class/leds/generic_%d/brightness", gpio);
	*fd = open(buf, O_WRONLY);
}

void gpio_set_value(int fd, int value)
{
    switch (value)
    {
        case 0: // LED on, GPIO pin low
            write(fd, "0", 1);
            break;
        case 1: // LED off, GPIO pin high
            write(fd, "1", 1);
            break;
        default:
            break;
    }
}

void gpio_close(int fd)
{
    close(fd);
}

void bitbang_string(int fd, const char* string)
{
	int i;
 
    for (i = 0; i < strlen(string); ++i)
    {
        bitbang_char(fd, string[i]);
        putchar (string[i]);        // Display character to console
        fflush(stdout);
        sleep(LOOP_DELAY);          // Inter-character delay
    }
}

void bitbang_char(int fd, char character)
{
    int i;
 
    // Start bit
    gpio_set_value(fd, 1); 
    udelay(BIT_TIME_US);

    // Data bits
    for (i = 0; i < 8; ++i) 
	{ 
        gpio_set_value(fd, !((character >> i) & 1)); 
        udelay(BIT_TIME_US);
    }
   
    // Stop bit
    gpio_set_value(fd, 0); 
    udelay(BIT_TIME_US);
}

void udelay(int us) // Poor man's delay (in microseconds)
{
    long i;

    for (i = 0; i < (us * 39); ++i);
}

